@@ -1,8 +1,13 @@
 
 \set ECHO none
 
+set client_min_messages = 'warning';
+
 \i plproxy.sql
 
+drop language if exists plpgsql;
+create language plpgsql;
+
 -- create cluster info functions
 create schema plproxy;
 create or replace function plproxy.get_cluster_version(cluster_name text)
@@ -11,6 +16,9 @@ begin
     if cluster_name = 'testcluster' then
         return 5;
     end if;
+    if cluster_name = 'badcluster' then
+        return 5;
+    end if;
     raise exception 'no such cluster: %', cluster_name;
 end; $$ language plpgsql;
 
@@ -22,6 +30,10 @@ begin
         return next 'host=127.0.0.1 dbname=test_part';
         return;
     end if;
+    if cluster_name = 'badcluster' then
+        return next 'host=127.0.0.1 dbname=nonex_db';
+        return;
+    end if;
     raise exception 'no such cluster: %', cluster_name;
 end; $$ language plpgsql;
 
@@ -36,28 +48,38 @@ end; $$ language plpgsql;
 -- intialize part
 -------------------------------------------------
 drop database if exists test_part;
-create database test_part;
-\c test_part
-create language plpgsql;
-
 drop database if exists test_part0;
-create database test_part0;
-\c test_part0
-create language plpgsql;
-
 drop database if exists test_part1;
-create database test_part1;
-\c test_part1
-create language plpgsql;
-
 drop database if exists test_part2;
-create database test_part2;
-\c test_part2
-create language plpgsql;
-
 drop database if exists test_part3;
+create database test_part;
+create database test_part0;
+create database test_part1;
+create database test_part2;
 create database test_part3;
+
+drop database if exists test_enc_proxy;
+drop database if exists test_enc_part;
+
+\c test_part
+set client_min_messages = 'warning';
+drop language if exists plpgsql;
+create language plpgsql;
+set client_min_messages = 'warning';
+\c test_part0
+set client_min_messages = 'warning';
+drop language if exists plpgsql;
+create language plpgsql;
+\c test_part1
+set client_min_messages = 'warning';
+drop language if exists plpgsql;
+create language plpgsql;
+\c test_part2
+set client_min_messages = 'warning';
+drop language if exists plpgsql;
+create language plpgsql;
 \c test_part3
+set client_min_messages = 'warning';
+drop language if exists plpgsql;
 create language plpgsql;
 
-
