@@ -117,7 +117,11 @@ as $$ cluster 'testcluster'; run on hashtext(username); $$ language plproxy;
 \c test_part
 create function test_types(username text, inout vbool boolean, inout xdate timestamp, inout bin bytea)
 as $$ begin return; end; $$ language plpgsql;
+
 \c regression
+select 1 from (select set_config(name, 'escape', false) as ignore
+          from pg_settings where name = 'bytea_output') x
+where x.ignore = 'foo';
 select * from test_types('types', true, '2009-11-04 12:12:02', E'a\\000\\001\\002b');
 select * from test_types('types', NULL, NULL, NULL);
 
@@ -143,6 +147,15 @@ create function test_connect1() returns text
 as $$ connect 'dbname=test_part'; select current_database(); $$ language plproxy;
 select * from test_connect1();
 
+-- test CONNECT $argument
+create function test_connect2(connstr text) returns text
+as $$ connect connstr; select current_database(); $$ language plproxy;
+select * from test_connect2('dbname=test_part');
+
+-- test CONNECT function($argument)
+create function test_connect3(connstr text) returns text
+as $$ connect text(connstr); select current_database(); $$ language plproxy;
+select * from test_connect3('dbname=test_part');
 
 -- test quoting function
 create type "RetWeird" as (
@@ -218,3 +231,48 @@ as $$
 $$ language plproxy;
 select * from test_simple(0);
 
+-- test error passing
+\c test_part
+create function test_error1() returns int4
+as $$
+begin
+    select line2err;
+    return 0;
+end;
+$$ language plpgsql;
+\c regression
+create function test_error1() returns int4
+as $$
+    cluster 'testcluster';
+    run on 0;
+$$ language plproxy;
+select * from test_error1();
+
+create function test_error2() returns int4
+as $$
+    cluster 'testcluster';
+    run on 0;
+    select err;
+$$ language plproxy;
+select * from test_error2();
+
+create function test_error3() returns int4
+as $$
+    connect 'dbname=test_part';
+$$ language plproxy;
+select * from test_error3();
+
+-- test invalid db
+create function test_bad_db() returns int4
+as $$
+    cluster 'badcluster';
+$$ language plproxy;
+select * from test_bad_db();
+
+create function test_bad_db2() returns int4
+as $$
+    connect 'dbname=wrong_name_db';
+$$ language plproxy;
+select * from test_bad_db2();
+
+
