@@ -29,6 +29,15 @@
 #include <fmgr.h>
 #include <executor/spi.h>
 
+#if PG_VERSION_NUM >= 80400
+#define PLPROXY_USE_SQLMED
+#include <foreign/foreign.h>
+#include <catalog/pg_foreign_data_wrapper.h>
+#include <catalog/pg_foreign_server.h>
+#include <catalog/pg_user_mapping.h>
+#endif
+
+#include <access/reloptions.h>
 #include <access/tupdesc.h>
 #include <catalog/pg_namespace.h>
 #include <catalog/pg_proc.h>
@@ -36,12 +45,15 @@
 #include <commands/trigger.h>
 #include <mb/pg_wchar.h>
 #include <miscadmin.h>
+#include <nodes/value.h>
+#include <utils/acl.h>
+#include <utils/array.h>
 #include <utils/builtins.h>
 #include <utils/hsearch.h>
+#include "utils/inval.h"
 #include <utils/lsyscache.h>
 #include <utils/memutils.h>
 #include <utils/syscache.h>
-
 #include "rowstamp.h"
 
 #include <libpq-fe.h>
@@ -63,6 +75,10 @@
 #define PG_DETOAST_DATUM_PACKED(x) PG_DETOAST_DATUM(x)
 #endif
 
+/*
+ * Determine if this argument is to SPLIT
+ */
+#define IS_SPLIT_ARG(func, arg)	((func)->split_args && (func)->split_args[arg])
 
 /*
  * Maintenece period in seconds.  Connnections will be freed
@@ -109,6 +125,7 @@ typedef struct ProxyConfig
 /* Single database connection */
 typedef struct
 {
+	struct ProxyCluster *cluster;
 	const char *connstr;		/* Connection string for libpq */
 
 	/* state */
@@ -118,9 +135,25 @@ typedef struct
 	ConnState	state;			/* Connection state */
 	time_t		connect_time;	/* When connection was started */
 	time_t		query_time;		/* When last query was sent */
-	bool		run_on;			/* True it this connection should be used */
 	bool		same_ver;		/* True if dest backend has same X.Y ver */
 	bool		tuning;			/* True if tuning query is running on conn */
+
+	/*
+	 * Nonzero if this connection should be used. The actual tag value is only
+	 * used by SPLIT processing, others should treat it as a boolean value.
+	 */
+	int			run_tag;
+
+	/*
+	 * Per-connection parameters. These are a assigned just before the 
+	 * remote call is made.
+	 */
+
+	Datum			   *split_params;					/* Split array parameters */
+	ArrayBuildState	  **bstate;							/* Temporary build state */
+	const char		   *param_values[FUNC_MAX_ARGS];	/* Parameter values */
+	int					param_lengths[FUNC_MAX_ARGS];	/* Parameter lengths (binary io) */
+	int					param_formats[FUNC_MAX_ARGS];	/* Parameter formats (binary io) */
 } ProxyConnection;
 
 /* Info about one cluster */
@@ -142,6 +175,20 @@ typedef struct ProxyCluster
 	int			ret_cur_conn;	/* Result walking: index of current conn */
 	int			ret_cur_pos;	/* Result walking: index of current row */
 	int			ret_total;		/* Result walking: total rows left */
+
+	bool		sqlmed_cluster;	/* True if the cluster is defined using SQL/MED */
+	bool		needs_reload;	/* True if the cluster partition list should be reloaded */
+	bool		busy;			/* True if the cluster is already involved in execution */
+
+	/*
+	 * SQL/MED clusters: TIDs of the foreign server and user mapping catalog tuples.
+	 * Used in to perform cluster invalidation in syscache callbacks.
+	 */
+	ItemPointerData		clusterTupleId;
+	ItemPointerData		umTupleId;
+
+	/* notice processing: provide info about currently executing function */
+	struct ProxyFunction	*cur_func;
 } ProxyCluster;
 
 /*
@@ -161,6 +208,10 @@ typedef struct ProxyType
 	bool		has_send;		/* Has binary output */
 	bool		has_recv;		/* Has binary input */
 	bool		by_value;		/* False if Datum is a pointer to data */
+	char		alignment;		/* Type alignment */
+	bool		is_array;		/* True if array */
+	Oid			elem_type;		/* Array element type */
+	short		length;			/* Type length */
 
 	/* I/O functions */
 	union
@@ -207,6 +258,17 @@ typedef struct ProxyQuery
 	void	   *plan;			/* Optional prepared plan for local queries */
 } ProxyQuery;
 
+/*
+ * Deconstructed array parameters
+ */
+typedef struct DatumArray
+{
+	ProxyType  *type;
+	Datum	   *values;
+	bool	   *nulls;
+	int			elem_count;
+} DatumArray;
+
 /*
  * Complete info about compiled function.
  *
@@ -224,6 +286,8 @@ typedef struct ProxyFunction
 	char	  **arg_names;		/* Argument names, may contain NULLs */
 	short		arg_count;		/* Argument count of proxy function */
 
+	bool	   *split_args;		/* Map of arguments to split */
+
 	/* if the function returns untyped RECORD that needs AS clause */
 	bool		dynamic_record;
 
@@ -239,6 +303,7 @@ typedef struct ProxyFunction
 	ProxyQuery *hash_sql;		/* Hash execution for R_HASH */
 	int			exact_nr;		/* Hash value for R_EXACT */
 	const char *connect_str;	/* libpq string for CONNECT function */
+	ProxyQuery *connect_sql;	/* Optional query for CONNECT function */
 
 	/*
 	 * calculated data
@@ -266,11 +331,15 @@ typedef struct ProxyFunction
 /* main.c */
 Datum		plproxy_call_handler(PG_FUNCTION_ARGS);
 void		plproxy_error(ProxyFunction *func, const char *fmt,...);
+void		plproxy_remote_error(ProxyFunction *func, ProxyConnection *conn, const PGresult *res, bool iserr);
 
 /* function.c */
 void		plproxy_function_cache_init(void);
 void	   *plproxy_func_alloc(ProxyFunction *func, int size);
 char	   *plproxy_func_strdup(ProxyFunction *func, const char *s);
+int			plproxy_get_parameter_index(ProxyFunction *func, const char *ident);
+bool		plproxy_split_add_ident(ProxyFunction *func, const char *ident);
+void		plproxy_split_all_arrays(ProxyFunction *func);
 ProxyFunction *plproxy_compile(FunctionCallInfo fcinfo, bool validate);
 
 /* execute.c */
@@ -299,6 +368,7 @@ void		plproxy_free_composite(ProxyComposite *meta);
 
 /* cluster.c */
 void		plproxy_cluster_cache_init(void);
+void		plproxy_syscache_callback_init(void);
 ProxyCluster *plproxy_find_cluster(ProxyFunction *func, FunctionCallInfo fcinfo);
 void		plproxy_cluster_maint(struct timeval * now);
 
@@ -311,8 +381,9 @@ bool		plproxy_query_add_const(QueryBuffer *q, const char *data);
 bool		plproxy_query_add_ident(QueryBuffer *q, const char *ident);
 ProxyQuery *plproxy_query_finish(QueryBuffer *q);
 ProxyQuery *plproxy_standard_query(ProxyFunction *func, bool add_types);
-void		plproxy_query_prepare(ProxyFunction *func, FunctionCallInfo fcinfo, ProxyQuery *q);
-void		plproxy_query_exec(ProxyFunction *func, FunctionCallInfo fcinfo, ProxyQuery *q);
+void		plproxy_query_prepare(ProxyFunction *func, FunctionCallInfo fcinfo, ProxyQuery *q, bool split_support);
+void		plproxy_query_exec(ProxyFunction *func, FunctionCallInfo fcinfo, ProxyQuery *q,
+							   DatumArray **array_params, int array_row);
 void		plproxy_query_freeplan(ProxyQuery *q);
 
 #endif
