@@ -253,6 +253,10 @@ plproxy_find_type_info(ProxyFunction *func, Oid oid, bool for_send)
 	type->for_send = for_send;
 	type->by_value = s_type->typbyval;
 	type->name = plproxy_func_strdup(func, namebuf);
+	type->is_array = (s_type->typelem != 0 && s_type->typlen == -1);
+	type->elem_type = s_type->typelem;
+	type->alignment = s_type->typalign;
+	type->length = s_type->typlen;
 
 	/* decide what function is needed */
 	if (for_send)
